/* WinNT zeitnet.h, Matt Welsh (mdw@cs.cornell.edu)
 *
 * Definitions for the U-Net/MM driver for the Zeitnet ATM NIC.
 *
 * M. Welsh, 18 Nov 96
 */

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)>(b)?(a):(b))

#define PCI_VENDOR_ID_ZEITNET 0x1193
#define PCI_DEVICE_ID_ZEITNET_1221 0x0001
#define PCI_DEVICE_ID_ZEITNET_1225 0x0002

static PHYSICAL_ADDRESS topPhysAddr = { 0xffffffff, 0 }; // end of phys mem

static void slow_down_io(void);

// Driver-defined settings
#define ZEITNET_MAX_DEVS 2
#define ZEITNET_MAX_ENDPT UNET_DEFAULT_MAX_ENDPOINTS
#define ZEITNET_MAX_CHANS 32
#define ZEITNET_MAX_PDU 65535 // AAL5 max

#define ZEITNET_MIN_TX 2
#define ZEITNET_MAX_TX 100
#define ZEITNET_MIN_FREE 2
#define ZEITNET_MAX_FREE 500
#define ZEITNET_MIN_RX 2
#define ZEITNET_MAX_RX 500

#define START_PREFETCH_THRESH 4  // TLB prefetch threshold
#define NUM_PREFETCH 8           // Number of buffers to prefetch
// Max number of pages in a receive buffer to prefetch. Used to define size
// of per-channel pool ring.
#define MAX_PREFETCH_BUFFER_PAGES  1
#define POOL_LIST_SIZE (2*(NUM_PREFETCH*MAX_PREFETCH_BUFFER_PAGES))

#define ZEITNET_MBX_RX_0_SIZE 256  // Mailbox sizes
#define ZEITNET_MBX_RX_1_SIZE 256
#define ZEITNET_MBX_TX_0_SIZE 256
#define ZEITNET_MBX_TX_1_SIZE 256

#define ZEITNET_NUM_POOLS 32
#define ZEITNET_NUM_SHAPERS 16

#define ZEITNET_TXRING_SIZE 32 // without link pointer

// Default traffic-shaping parameters
#define ZEITNET_DEFAULT_I 5
#define ZEITNET_DEFAULT_M 10
#define ZEITNET_DEFAULT_P 0
#define ZEITNET_DEFAULT_C 5

// register offsets
#define ZN_GMR 0x00
#define ZN_GSR 0x04
#define ZN_IMR 0x08
#define ZN_RQU 0x0c
#define ZN_RQA 0x10
#define ZN_ADDR 0x14
#define ZN_VER 0x18
#define ZN_SWR 0x1c
#define ZN_CMR 0x20
#define ZN_CMR_L 0x24
#define ZN_CER 0x28
#define ZN_CER_L 0x2c

#define ZN_MSH(i) (0x40 + (i*4))
#define ZN_MSL(i) (0x50 + (i*4))
#define ZN_MBA(i) (0x60 + (i*4))
#define ZN_MTA(i) (0x70 + (i*4))
#define ZN_MWA(i) (0x80 + (i*4))

// Indirectly-accessible registers
#define ZN_SCHED_IM 0x40000
#define ZN_SCHED_X  0x40010
#define ZN_SCHED_Y  0x40020
#define ZN_SCHED_PCpc 0x40030
#define ZN_SCHED_pristat 0x40040
#define ZN_TOS 0x40100
#define ZN_SMA 0x40200
#define ZN_PMA 0x40201
#define ZN_T1R 0x40300
#define ZN_VRR 0x40301
#define ZN_TSR 0x40302

// Register values
#define ZN_CMR_BUSY 0x80000000

// Misc defines
#define ZN_MAX_CRAM_SIZE (1<<18)
#define ZN_RX_LOOKUP_SIZE 2    // in bytes
#define ZN_POOL_DESCR_SIZE 8   // in bytes
#define ZN_SHAPER_SIZE 4       // in bytes
#define ZN_VC_DESCR_SIZE 32    // in bytes

#define ZN_NUM_MBX  4 // mailboxes
#define ZN_MBX_RX_0 0
#define ZN_MBX_RX_1 1
#define ZN_MBX_TX_0 2
#define ZN_MBX_TX_1 3

// Utility macros
#define outl(data, addr) WRITE_PORT_ULONG((unsigned long *)(addr), data)
#define inl(addr) READ_PORT_ULONG((unsigned long *)(addr))

#define zin(reg) inl(zeitnet->cbio+ZN_##reg)
#define zout(val, reg) outl(val, zeitnet->cbio+ZN_##reg)
#define zwait(_x) { while(zin(CMR) & ZN_CMR_BUSY) { if ((_x)) slow_down_io(); }}

/* ZN/uPD98401 data structures *********************************************/

typedef struct {
	USHORT pkt_size;
	USHORT uinfo;
	ULONG timestmp;
	ULONG addr;
	unsigned int pool:5;
	unsigned int pad1:3;
	unsigned int err_status:4;
	unsigned int pad2:1;
	unsigned int clp:1;
	unsigned int ci:1;
	unsigned int err:1;
	unsigned int conn:15;
	unsigned int valid:1;
} ZN_RX_MBX_ENTRY;
#define ZN_RX_MBX_SIZE (sizeof(ZN_RX_MBX_ENTRY))

typedef struct {
	unsigned int pkt_queue:15;
	unsigned int active:1;
	unsigned int conn:15;
	unsigned int valid:1;
} ZN_TX_MBX_ENTRY;
#define ZN_TX_MBX_SIZE (sizeof(ZN_TX_MBX_ENTRY))

typedef struct {
	ULONG status;
	ULONG pad;
	ULONG size;
	ULONG addr;
} ZN_TX_DESCR;
#define ZN_TX_DESCR_SIZE (sizeof(ZN_TX_DESCR))

typedef struct {
	ULONG size;
	ULONG addr;
} ZN_TX_PKTDIR;
#define ZN_TX_PKTDIR_SIZE (sizeof(ZN_TX_PKTDIR))

/* Zeitnet data structures **************************************************/

// Channel tx_ring shadow
typedef struct {
	ULONG page_id;
	ULONG paddr;
} ZEITNET_TX_SHADOW;

typedef struct {
	BOOLEAN last;
	ZEITNET_TX_SHADOW shadow;
	ZN_TX_PKTDIR *pktdir; 
} ZEITNET_PKTDIR_SHADOW;

// Free pool
typedef struct _ZEITNET_FREE_POOL {
	ULONG paddr;  // Physical address, used by SAR
	ULONG p_next; // Physical batch link ptr, used by SAR
	ULONG vaddr;
	ULONG saddr;
} ZEITNET_FREE_POOL, *PZEITNET_FREE_POOL;

// Traffic shaper
typedef struct _ZEITNET_SHAPER {
	int refCnt;
	UCHAR I, P, C;
	ULONG M;
} ZEITNET_SHAPER, *PZEITNET_SHAPER;

// Channel
typedef struct {
  int inUse;
	int index; // Channel number
  UNET_ATM_ADDRESS	txAddr;		// Tx destination address
  UNET_ATM_ADDRESS	rxAddr;		// Rx destination address
	struct ZEITNET_ENDPT *endpt;    // Back pointer to endpt

	int tx_conn, rx_conn;
	int tx_shaper;
	ZN_TX_DESCR *tx_ring, *tx_ring_next, *tx_ring_last;
	volatile ZN_TX_DESCR *tx_ring_dirty;
	ZEITNET_TX_SHADOW *tx_ring_shadow;
	ULONG tx_ring_phys; 
	volatile BOOLEAN tx_ring_full;
	
} ZEITNET_CHAN, *PZEITNET_CHAN;

// Device
typedef struct {
  // Device addresses and NT resource handles
  int index;                            // device index
  ULONG cbio;		                        // I/O base addr for CSRs
	ULONG cbma;                           // memory base addr for CSRs
  unsigned char macAddr[6];		          // MAC address from EEPROM
  PUNICODE_STRING registryPath;		      // used for releasing resources
  PKINTERRUPT interruptObject;
  PDEVICE_OBJECT deviceObject;

	BOOLEAN adapter_created;

	BOOLEAN copper;

	// Control memory size and offsets
	ULONG memsz;
	ULONG pool_base;

	// Mailboxes
	PULONG mbx[ZN_NUM_MBX], mbx_next[ZN_NUM_MBX], mbx_dirty[ZN_NUM_MBX], mbx_end[ZN_NUM_MBX];
	ULONG p_mbx[ZN_NUM_MBX];
	// Mapping from transmit connections to channels (for reap)
	int max_conns;
	PZEITNET_CHAN *tx_conn_map, *rx_conn_map;

	// Free buffer pools usage list
	BOOLEAN pool_inuse[ZEITNET_NUM_POOLS];
	// Shapers
	ZEITNET_SHAPER shaper[ZEITNET_NUM_SHAPERS];

} ZEITNET_DEV, *PZEITNET_DEV;

// Endpoint
typedef struct ZEITNET_ENDPT {
  PZEITNET_DEV	dev;
	PEPROCESS     process;
	int index;
	
  UNET_TX_BUFFER_DESCRIPTOR		*txFirst;
  volatile UNET_TX_BUFFER_DESCRIPTOR	*txNext;
  UNET_TX_BUFFER_DESCRIPTOR		*txLast;
	UNET_TX_BUFFER_DESCRIPTOR   *txDirty; // Corresponds to wall
  ULONG					*txWall;
  int					numTxFifoEntries;

  UNET_RECV_BUFFER_DESCRIPTOR	       *freeFirst;
  volatile UNET_RECV_BUFFER_DESCRIPTOR *freeNext;
  UNET_RECV_BUFFER_DESCRIPTOR	       *freeLast;
  ULONG					*freeWall;
  int					numFreeFifoEntries;

	// For prefetching of free buffers
  volatile UNET_RECV_BUFFER_DESCRIPTOR *freePrefetchNext;
	volatile int freePrefetch;

  UNET_RECEIVE_DESCRIPTOR		*rxFirst;
  volatile UNET_RECEIVE_DESCRIPTOR	*rxNext;
  UNET_RECEIVE_DESCRIPTOR		*rxLast;
  ULONG					*rxWall;
  int					numRxFifoEntries;

  ULONG					fifoAreaSize;
  ULONG					bufferAreaSize;
  void					*bufferArea;
  ULONG					userBufferArea;
  ULONG					userFifoArea;
  ULONG					physBufferArea;
	int           rxBufferSize, rxBufferUsedSize;
	int           rxNumCells;

  int					numChans;
  ZEITNET_CHAN				chans[ZEITNET_MAX_CHANS];

	int num_freed; // Number of Tx fifo entries freed during reap

	int pool_num;
	ZEITNET_FREE_POOL pool[POOL_LIST_SIZE],
		*pool_next, *pool_dirty, *pool_last;

	// Currently-used shaper
	int tx_shaper;

} ZEITNET_ENDPT, *PZEITNET_ENDPT;
