/* libunet.c, M. Welsh (mdw@cs.cornell.edu)
 *
 * Copyright (c) 1995, 1996 by Cornell University.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL CORNELL UNIVERSITY BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF CORNELL UNIVERSITY HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * CORNELL UNIVERSITY SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND CORNELL UNIVERSITY HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Derived from libunet.c, v1.13
 * Ported to NT4.0 by mdw, 23-09-96.
 * Updated for U-Net/MM extensions, mdw, 21-10-96.
 *
 */

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unetuser.h>

#include "libunet.h"

#undef LIBUNET_DEBUG   /* Define for verbose debugging */
#define LIBUNET_LOG    /* Define for error conditions to be printed */

#ifndef PENTIUM_MHZ

/* Non-Pentium NT-friendly version */
double get_seconds(void) {
  DWORD t;
  t = GetTickCount();
  return (double)t/(double)1e3;
}

#else

/* Pentium version --- PENTIUM_MHZ defined to MHz rating of processor */
double get_seconds() {
  unsigned long f1, f2;
  __int64 now;
  __asm {
    __asm _emit 0x0f
    __asm _emit 0x31
    __asm mov f1, eax
    __asm mov f2, edx
  }  
  now = f1 | ((__int64)f2 << 32);
  return (double)now / (double)(1e6*PENTIUM_MHZ);
}

#endif /* PENTIUM_MHZ */

int str_to_device(char *str, UNET_DEV_TYPE *devtype, UNET_DEV_INDEX *devindex) {
  char *dname, *iname;
  dname = strtok(str, "/");
  iname = strtok(NULL, "/");

  if (!dname || !iname) return -1;

  if (!strcmp(dname, "sba200")) {
    *devtype = UNET_DEVID_SBA200;
  } else if (!strcmp(dname, "pca200")) {
    *devtype = UNET_DEVID_PCA200;
  } else if (!strcmp(dname, "tulip")) {
    *devtype = UNET_DEVID_TULIP;
  } else if (!strcmp(dname, "zeitnet")) {
    *devtype = UNET_DEVID_ZEITNET;
  } else {
    return -1;
  }
  *devindex = atoi(iname);
  return 0;
}


/* Convert a string into a unet_chan_addr structure */
int str_to_addr(UNET_DEV_TYPE devtype, char *str, UNET_CHANNEL_ADDRESS *addr) {
  if (devtype == UNET_DEVID_TULIP) {
    unsigned int mac_addr[6];
    unsigned int port;
    if (sscanf(str,"%x:%x:%x:%x:%x:%x.%x",
	   &(mac_addr[0]),
	   &(mac_addr[1]),
	   &(mac_addr[2]),
	   &(mac_addr[3]),
	   &(mac_addr[4]),
	   &(mac_addr[5]),
	   &port) != 7) return -1;

    addr->fe.macAddr[0] = (unsigned char)mac_addr[0];
    addr->fe.macAddr[1] = (unsigned char)mac_addr[1];
    addr->fe.macAddr[2] = (unsigned char)mac_addr[2];
    addr->fe.macAddr[3] = (unsigned char)mac_addr[3];
    addr->fe.macAddr[4] = (unsigned char)mac_addr[4];
    addr->fe.macAddr[5] = (unsigned char)mac_addr[5];
    addr->fe.port = (unsigned char)port;
    return 0;
  }

  if ((devtype == UNET_DEVID_SBA200) || (devtype == UNET_DEVID_PCA200) ||
       (devtype == UNET_DEVID_ZEITNET)) {
    if (sscanf(str,"%d",&(addr->atm.vci)) != 1) return -1;
    return 0;
  }

  return -1;
}

HANDLE unet_opendev(UNET_DEV_TYPE devtype, UNET_DEV_INDEX devindex,
		    UNET_INFO_RESPONSE *info) {

  HANDLE hDriver;
  int nadapters, nbytes, i;
  UNET_INFO_RESPONSE *adapterlist;

  hDriver = CreateFile("\\\\.\\UNET", GENERIC_READ | GENERIC_WRITE,
		       0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
		       NULL);
  if (hDriver == (HANDLE)-1) {
    Perror("unet_opendev: CreateFile");
  }

  if (!DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_ADAPTERS, NULL, 0,
		       &nadapters, sizeof(ULONG), &nbytes, 0)) {
    Perror("IOCTL_UNET_ADAPTERS");
    CloseHandle(hDriver);
    return (HANDLE)NULL;
  }

  adapterlist = (UNET_INFO_RESPONSE *)malloc(nadapters * sizeof(UNET_INFO_RESPONSE));
  if (!adapterlist) return (HANDLE)NULL;

  if (!DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_ADAPTER_LIST, NULL, 0,
		       adapterlist, nadapters * sizeof(UNET_INFO_RESPONSE),
		       &nbytes, 0)) {
    Perror("IOCTL_UNET_ADAPTER_LIST");
    CloseHandle(hDriver);
    return (HANDLE)NULL;
  }

  for (i = 0; i < nadapters; i++) {
    if ((adapterlist[i].dev_type == devtype) &&
	(adapterlist[i].dev_index == devindex)) {
      *info = adapterlist[i];
      free(adapterlist);
      return hDriver;
    }
  }

  free(adapterlist);
  CloseHandle(hDriver);
  return (HANDLE)NULL;
}
    
/* Create an endpoint on the given (open) U-Net device */
struct unet_endpt_user *unet_create_endpt(HANDLE hDriver,
					  UNET_INFO_RESPONSE *info,
					  UNET_CREATE_ENDPOINT_REQUEST *create,
					  int num_rx_buffers) {
 
  struct unet_endpt_user *endpt;
  UNET_CREATE_ENDPOINT_RESPONSE response;
  caddr_t comseg;
  int i, nbytes;

  if (!DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_CREATE_ENDPOINT,
		       create, sizeof(UNET_CREATE_ENDPOINT_REQUEST),
		       &response, sizeof(response),
		       &nbytes, 0)) {
    Perror("IOCTL_UNET_CREATE_ENDPOINT");
    return NULL;
  }

  /* Create endpoint struct */
  endpt = (struct unet_endpt_user *)malloc(sizeof(struct unet_endpt_user));
  if (endpt == NULL) {
    CloseHandle(hDriver);  /* XXX need to destroy endpoint here */
    return NULL;
  }

  endpt->endpointId = response.endpointId;
  endpt->hDriver = hDriver;
  endpt->info = *info;

  endpt->tx_first = response.txFifo;
  endpt->tx_next = endpt->tx_first;
  endpt->tx_last = endpt->tx_first + (create->numTxFifoEntries - 1);
  endpt->tx_wall = response.txWall;
  endpt->tx_wall_cache = (UNET_TX_BUFFER_DESCRIPTOR *)((u_int)endpt->tx_first +
						       *(endpt->tx_wall));
 
  endpt->free_first = response.freeFifo;
  endpt->free_next = endpt->free_first;
  endpt->free_last = endpt->free_first + (create->numFreeFifoEntries - 1);
  endpt->free_wall = response.freeWall;
 
  endpt->rx_first = response.rxFifo;
  endpt->rx_next = endpt->rx_first;
  endpt->rx_last = endpt->rx_first + (create->numRxFifoEntries - 1);
  endpt->rx_wall = response.rxWall;

  endpt->buffer_area = response.bufferArea;
#if 0 /* mdw: Only if Rx buffers always static */
  endpt->rxb_area = endpt->buffer_area;
#else
#define PAGE_SIZE 0x1000 // XXX mdw: Should be defined by NT
	endpt->rxb_area = malloc(num_rx_buffers * response.rxBufferAllocSize + PAGE_SIZE);
	endpt->rxb_area = ((unsigned long)endpt->rxb_area + (PAGE_SIZE-1)) & ~(PAGE_SIZE-1);
#endif
  endpt->rxb_numbuff = num_rx_buffers;
  endpt->rxb_buffsize_alloc = response.rxBufferAllocSize;
	endpt->rxb_buffsize_used  = response.rxBufferUsedSize;
  endpt->txb_area = endpt->buffer_area + (num_rx_buffers * response.rxBufferAllocSize);

  /* Set up the Rx FIFO */
  for (i = 0; i < create->numRxFifoEntries; i++) {
    (endpt->rx_first + i)->chan = UNET_CHANNEL_FREE;
  }
 
  /* Set up the free FIFO. Note that we only push on as many entries as
   * we have Rx buffers for: Also important is that free_next is incremented
   * as we do so. This means that new entries are pushed on starting AFTER
   * these initial entries (if there are more free FIFO entries than Rx
   * buffers).
   */

  /* Check that the Rx buffers will be properly aligned for the device */
  if (create->rxBufferSize & (endpt->info.rx_buf_align - 1)) {
    CloseHandle(hDriver);  /* XXX Should destroy endpoint here */
    free(endpt);
    return(NULL);
  }

	/* Put virtual addresses of receive buffers in the free FIFO */
  for (i = 0; i < MIN(create->numFreeFifoEntries, num_rx_buffers); i++) {
		endpt->free_next->buf = endpt->rxb_area + (i * response.rxBufferAllocSize);
    if (endpt->free_next == endpt->free_last) 
      endpt->free_next = endpt->free_first;
    else
      endpt->free_next++;
  }
 
  /* Initialize the Rx wall pointer */
  *(endpt->rx_wall) = (u_int)endpt->rx_last - (u_int)endpt->rx_first;

  endpt->rxwall_deferred = 0;

  return endpt;
}

/* Destroy the endpoint on the (given) open U-Net device */
int unet_destroy_endpt(HANDLE hDriver, UNET_ENDPOINT_ID endpointId) {

  UNET_DESTROY_ENDPOINT_REQUEST destroy;
  int ret, nbytes;

  destroy.endpointId = endpointId;
  
  if (!(ret = DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_DESTROY_ENDPOINT,
			     &destroy, sizeof(UNET_DESTROY_ENDPOINT_REQUEST),
			      NULL, 0, &nbytes, 0))) {
    Perror("IOCTL_UNET_DESTROY_ENDPT");
    return ret;
  }
  return ret;
}
 
/* Activate a channel on the given U-Net device */
UNET_CHANNEL_ID unet_activate_chan(HANDLE hDriver,
				   UNET_ACTIVATE_CHANNEL_REQUEST *activate) {

  int nbytes;
  UNET_ACTIVATE_CHANNEL_RESPONSE response;
  if (!DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_ACTIVATE_CHANNEL,
		       activate, sizeof(UNET_ACTIVATE_CHANNEL_REQUEST),
		       &response, sizeof(response),
		       &nbytes, 0)) {
    Perror("IOCTL_UNET_ACTIVATE_CHANNEL");
    return (UNET_CHANNEL_ID)-1;
  }
  return response.chan;
}

/* Deactivate a channel on the given U-Net device */
int unet_deactivate_chan(HANDLE hDriver,
			 UNET_DEACTIVATE_CHANNEL_REQUEST *deactivate) {

  int nbytes, ret;
  if (!(ret = DeviceIoControl(hDriver, (DWORD)IOCTL_UNET_DEACTIVATE_CHANNEL,
			      deactivate,
			      sizeof(UNET_DEACTIVATE_CHANNEL_REQUEST),
			      NULL, 0, &nbytes, 0))) {
    Perror("IOCTL_UNET_DEACTIVATE_CHANNEL");
    return ret;
  }
  return ret;
}

void unet_update_txwall(struct unet_endpt_user *endpt) {
  
#if (defined(UNET_DEV_TULIP) || defined(UNET_DEVID_ZEITNET))
  /* XXX Should be a run-time test */
  UNET_TRAP(endpt->hDriver, endpt->endpointId);
#endif
  endpt->tx_wall_cache = (UNET_TX_BUFFER_DESCRIPTOR *)((u_int)endpt->tx_first +
						       *(endpt->tx_wall));
}

/* Add a single buffer to the Tx FIFO */
int unet_send(struct unet_endpt_user *endpt, UNET_CHANNEL_ID chan,
	      unsigned char *buf, int len) {

  UNET_TX_BUFFER_DESCRIPTOR txbd;
  txbd.buf = (unsigned long)buf;
  txbd.chan = chan;
  txbd.length = len;

  /* Try to get free Tx FIFO entry */
  if (endpt->tx_next == endpt->tx_wall_cache) {
    unet_update_txwall(endpt);
    if (endpt->tx_next == endpt->tx_wall_cache) {
      /* Tx FIFO is full! */
      return -1;
    }
  }

#ifdef LIBUNET_DEBUG  
  fprintf(stderr,"buf: 0x%x, len: 0x%x, chan: 0x%x\n",
	  ((unsigned long)buf, len, chan);
#endif

  *(endpt->tx_next) = txbd;

#if (defined(UNET_DEV_TULIP) || defined(UNET_DEVID_ZEITNET))
  UNET_TRAP(endpt->hDriver, endpt->endpointId);
#endif

  if (endpt->tx_next == endpt->tx_last)
    endpt->tx_next = endpt->tx_first;
  else
    endpt->tx_next++;

  return 0;
}

int unet_sendv(struct unet_endpt_user *endpt, UNET_CHANNEL_ID chan,
	       struct iovec *iov, int iovcnt) {
  UNET_TX_BUFFER_DESCRIPTOR txbd;
  int i;

  /* Try to get free Tx FIFO entry */
  if (endpt->tx_next == endpt->tx_wall_cache) {
    unet_update_txwall(endpt);
    if (endpt->tx_next == endpt->tx_wall_cache) {
      /* Tx FIFO is full! */
      return -1;
    }
  }

  /* Enough space in the FIFO for all these buffers? */
  if (endpt->tx_wall_cache > endpt->tx_next) {
    if ((endpt->tx_wall_cache - endpt->tx_next) <= iovcnt) {
      unet_update_txwall(endpt);
      if (endpt->tx_wall_cache > endpt->tx_next) {
	if ((endpt->tx_wall_cache - endpt->tx_next) <= iovcnt) {
	  return -1;
	}
      } else {
	if (((endpt->tx_last - endpt->tx_next) + (endpt->tx_wall_cache -
						  endpt->tx_first)) <= iovcnt) {
	  return -1;
	}
      }
    }
  } else {
    if (((endpt->tx_last - endpt->tx_next) + (endpt->tx_wall_cache -
					      endpt->tx_first)) <= iovcnt) {
      unet_update_txwall(endpt);
      if (endpt->tx_wall_cache > endpt->tx_next) {
	if ((endpt->tx_wall_cache - endpt->tx_next) <= iovcnt) {
	  return -1;
	}
      } else {
	if (((endpt->tx_last - endpt->tx_first) + (endpt->tx_wall_cache -
						   endpt->tx_first)) <= iovcnt) {
	  return -1;
	}
      }
    }
  }

  txbd.chan = chan;

  /* Send all but the last buffer with the continuation bit set */
  iovcnt--;
  for (i = 0; i < iovcnt; i++) {
    txbd.buf = ((unsigned long)iov[i].iov_base) | 0x01;
    txbd.length = iov[i].iov_len;

    *(endpt->tx_next) = txbd;
    if (endpt->tx_next == endpt->tx_last) {
      endpt->tx_next = endpt->tx_first;
    } else {
      endpt->tx_next++;
    }
  }

  /* Send last without continuation bit */
  txbd.buf = ((unsigned long)iov[i].iov_base);
  txbd.length = iov[i].iov_len;

  *(endpt->tx_next) = txbd;

#if (defined(UNET_DEV_TULIP) || defined(UNET_DEV_ZEITNET))
  UNET_TRAP(endpt->hDriver, endpt->endpointId);
#endif
  
  if (endpt->tx_next == endpt->tx_last) {
    endpt->tx_next = endpt->tx_first;
  } else {
    endpt->tx_next++;
  }
  
  return 0;
}


/* Spin until a new Rx FIFO entry arrives */
int unet_recv_poll(struct unet_endpt_user *endpt) {

  while (endpt->rx_next->chan == UNET_CHANNEL_FREE) ;

  /* Skip over corrupted messages and re-poll */
  while (endpt->rx_next->chan == UNET_CHANNEL_ERROR) {
    /* XXX mdw (26 Mar 96): This needs to push whatever buffers were
     * allocated in the rxd back onto the free FIFO, else they're
     * lost forever...
     */

#ifdef LIBUNET_LOG
    fprintf(stderr,"unet_recv_poll: Error on receive, Unet_rx len %d\n",
       endpt->rx_next->length);
#endif

    endpt->rx_next->chan = UNET_CHANNEL_FREE;
    *(endpt->rx_wall) = (u_int)endpt->rx_next - (u_int)endpt->rx_first;
       
    if (endpt->rx_next == endpt->rx_last)
      endpt->rx_next = endpt->rx_first;
    else
      endpt->rx_next++;

    while (endpt->rx_next->chan == UNET_CHANNEL_FREE) ;
  }

  return endpt->rx_next->chan;
}

/* Retrieve a message from the Rx FIFO in iov format */
int unet_recv(struct unet_endpt_user *endpt, UNET_CHANNEL_ID *chan,
	      struct iovec *iov, int *iovcnt) {
  int i, l, len;
  
  if (endpt->rxwall_deferred) {
    /* rx_wall not updated last time around because the message fit into
     * the Rx descriptor. Here we assume that the next time into
     * unet_recv, the user is done with the data in the descriptor and
     * we can push the wall past it.
     */
    endpt->rxwall_deferred_entry->chan = UNET_CHANNEL_FREE;
    *(endpt->rx_wall) = (u_int)endpt->rxwall_deferred_entry - (u_int)endpt->rx_first;
    endpt->rxwall_deferred = 0;
  }

  /* First skip any corrupted messages */
  while (endpt->rx_next->chan == UNET_CHANNEL_ERROR) {

#ifdef LIBUNET_LOG
    fprintf(stderr,"unet_recv: Error on receive, Unet_rx len %d\n",
       endpt->rx_next->length);
#endif

    endpt->rx_next->chan = UNET_CHANNEL_FREE;
    *(endpt->rx_wall) = (u_int)endpt->rx_next - (u_int)endpt->rx_first;

    if (endpt->rx_next == endpt->rx_last)
      endpt->rx_next = endpt->rx_first;
    else
      endpt->rx_next++;

  }

  if (endpt->rx_next->chan == UNET_CHANNEL_FREE) {
    /* No data! */
    return -1;
  }

  *chan = endpt->rx_next->chan;
  l = len = endpt->rx_next->length;

#ifdef LIBUNET_DEBUG
  fprintf(stderr,"Got message, chan %d, len %d\n",*chan,l);
#endif

  if (l > endpt->info.rxd_payload_size) {
    i = 0;
    while (l > 0) {
      iov[i].iov_base = (caddr_t) (endpt->rx_next->pay_or_buf[i]);
      iov[i].iov_len = MIN(l, endpt->rxb_buffsize_used);

#ifdef LIBUNET_DEBUG
      fprintf(stderr,"iov[%d].iov_base: 0x%p, iov_len: %d\n",
      	i,iov[i].iov_base,iov[i].iov_len);
      {
        int j;
	for (j = 0; j < iov[i].iov_len; j++) {
	  fprintf(stderr,"%2.2x ",*((unsigned char *)iov[i].iov_base + j));
 	}
	fprintf(stderr,"\n");
      }
#endif

      l -= iov[i].iov_len;
      i++;
    }
    *iovcnt = i;
    endpt->rx_next->chan = UNET_CHANNEL_FREE;
    *(endpt->rx_wall) = (u_int)endpt->rx_next - (u_int)endpt->rx_first;
    
  } else {
    iov[0].iov_base = (caddr_t)endpt->rx_next->pay_or_buf;
    iov[0].iov_len = l;
    *iovcnt = 1;

    /* Don't want to push rx_wall until the user is 'done' with this
     * entry... (Assume that to be next time into this function).
     */
    endpt->rxwall_deferred = 1; 
    endpt->rxwall_deferred_entry = endpt->rx_next;
  }

  /* Bump rx_next */
  if (endpt->rx_next == endpt->rx_last)
    endpt->rx_next = endpt->rx_first;
  else
    endpt->rx_next++;

  return len;
}

/* Buffer management routines */

/* Push a buffer onto the free FIFO. Assumes that addr is an offset into
 * the buffer area.
 */
void unet_free_rxbuf(struct unet_endpt_user *endpt,
		     unsigned int addr) {
  
  endpt->free_next->buf = addr;
  if (endpt->free_next == endpt->free_last)
    endpt->free_next = endpt->free_first;
  else
    endpt->free_next++;
}
  
/* Given an iov, push all buffers onto the free fifo. Assumes that
 * addresses in the iov are within the buffer area.
 */
void unet_free_rxv(struct unet_endpt_user *endpt, struct iovec *iov, int iovc) {
  int i;

  /* Inlined message */
  if (iov[0].iov_len <= endpt->info.rxd_payload_size)
    return;

  for (i = 0; i < iovc; i++)
    unet_free_rxbuf(endpt, (unsigned int)iov[i].iov_base);

  return;
}

#if 0   /* XXX mdw (23-Sep-96): As yet unimplemented for NT4.0 */

/* Signal/notification functions */
int unet_intr(int fd, void(*handler)()) {
  int on = 1;
  
  /* XXX mdw: Should use sigaction */
  signal(SIGIO, handler);
  if (ioctl(fd, FIOASYNC, &on)) return -1;
  else return 0;
}

int unet_notify(int fd, Unet_intr_ctrl *intrctrl) {
  if (ioctl(fd, UNET_INTR, intrctrl)) return -1;
  else return 0;
}

#endif /* 0 from above */
