/* loopback.c, M. Welsh (mdw@cs.cornell.edu)
 *
 * A simple loop-back test program for U-Net/NT4.0.
 *
 * This version modified for U-Net/MM, mdw.
 */

#undef MDW_DEBUG
#undef MDW_TERSE_OUTPUT
#undef MDW_TIME_TRAPS
#undef MDW_TEST_IOCTL

#include <windows.h>
#include <winioctl.h>
#include <stdio.h>
#include <stdlib.h>

#include "unetuser.h"
#include "libunet.h"

#ifdef MDW_TEST_IOCTL
#include "zeitnetuser.h"
#endif

#define TX_FIFO_SIZE 10

unsigned char thebuf[8192];

int __cdecl main(int  argc, char *argv[]) {
  HANDLE      hDriver;
  UNET_INFO_RESPONSE info;
  int message_size, num_messages;
  struct unet_endpt_user *endpt;
  UNET_CREATE_ENDPOINT_REQUEST create;
  UNET_ACTIVATE_CHANNEL_REQUEST activate;
  UNET_CHANNEL_ID chan;
  UNET_TX_BUFFER_DESCRIPTOR txbd;
  double t1, t2;
  int i, j, n;
  char *addr, *devstr;
  UNET_DEV_TYPE devtype; 
  UNET_DEV_INDEX devindex;
  int pdu_size; 
  int nBytes;
  int num_bufs, len;
#ifdef MDW_TEST_IOCTL
  UNET_ADAPTER_IOCTL_REQUEST ioctl_req;
  UNET_ADAPTER_IOCTL_RESPONSE ioctl_resp;
  UNET_ZEITNET_IOCTL_TRAFFICSHAPING_REQUEST treq;
  UNET_ZEITNET_IOCTL_TRAFFICSHAPING_RESPONSE tresp;
#endif

  if (argc != 5) {
    fprintf(stderr,"usage: loopback <device> <address> <num> <size>\n");
    exit(-1);
  }
  
  devstr = argv[1];
  addr = argv[2];
  num_messages = atoi(argv[3]);
  message_size = atoi(argv[4]);

  if (str_to_device(devstr, &devtype, &devindex)) {
    fprintf(stderr,"Device string not in format <devname>.<index>.\n");
    exit(-1);
  }

  hDriver = unet_opendev(devtype, devindex, &info);
  if (!hDriver) {
    fprintf(stderr,"Can't open U-Net device!\n");
    exit(-1);
  }
  pdu_size = MIN(4096, info.max_pdu);

  // Create an endpoint
  create.devType = devtype;
  create.devIndex = devindex;
  create.numTxFifoEntries   = TX_FIFO_SIZE;
  create.numFreeFifoEntries = TX_FIFO_SIZE;
  create.numRxFifoEntries   = TX_FIFO_SIZE;
  create.bufferAreaSize     = (TX_FIFO_SIZE+1)*pdu_size;
  create.rxBufferSize       = pdu_size;

  endpt = unet_create_endpt(hDriver, &info, &create, TX_FIFO_SIZE);
  if (!endpt) {
    fprintf(stderr,"Can't create endpoint!\n");
    exit(-1);
  }
  
  // Open a channel
  activate.endpointId = endpt->endpointId;
  activate.chan = UNET_CHAN_ANY;
  if (str_to_addr(devtype, addr, &(activate.txAddr))) {
    fprintf(stderr,"Bad address format %s.\n",addr);
    exit(-1);
  }
  (void)str_to_addr(devtype, addr, &(activate.rxAddr));

  chan = unet_activate_chan(hDriver, &activate);
  if (chan == -1) {
    fprintf(stderr,"Can't activate channel!\n");
    exit(-1);
  }

#ifdef MDW_TEST_IOCTL
  /* Test ADAPTER_IOCTL */
  ioctl_req.endpointId = endpt->endpointId;
  ioctl_req.code = UNET_ZEITNET_IOCTL_TRAFFICSHAPING;
  ioctl_req.bufferSize = sizeof(treq);
  ioctl_req.buffer = &treq;
  ioctl_resp.bufferSize = sizeof(tresp);
  ioctl_resp.buffer = &tresp;

  treq.I = 8;
  treq.M = 10;
  treq.P = 4;
  treq.C = 5;

  if (!DeviceIoControl(hDriver, IOCTL_UNET_ADAPTER_IOCTL,
                       &ioctl_req, sizeof(ioctl_req),
		       &ioctl_resp, sizeof(ioctl_resp),
		       &nBytes, 0)) {
    fprintf(stderr,"Can't call IOCTL_UNET_ADAPTER_IOCTL.\n");
    exit(-1);
  }

#ifndef MDW_TERSE_OUTPUT
  fprintf(stderr,"Got ioctl response %d %d %d %d.\n",
    tresp.I, tresp.M, tresp.P, tresp.C);
#endif

#endif
		         
  txbd.buf = thebuf;
  txbd.length = message_size;
  txbd.chan = chan;
	
  for (i = 0; i < message_size; i++) {
    thebuf[i] = i&0xff;
  }

#ifdef MDW_TIME_TRAPS
  {
    int i;
    t1 = get_seconds();
    for (i = 0; i < 1000; i++) {
      UNET_TRAP(hDriver, endpt->endpointId);
    }
    t2 = get_seconds();
    fprintf(stderr,"1000 traps cost %f secs or %f usec/trap\n",
	    t2-t1, ((t2-t1)*1e6)/1000);
  }
#endif

  /* Do a lot of them */
#ifndef MDW_TERSE_OUTPUT
  fprintf(stderr,"Sending %d %d-byte messages.\n",
	  num_messages, message_size);
#endif
  
  t1 = get_seconds();
  
  for (i = 0; i < num_messages; i++) {
#ifdef MDW_DEBUG
    fprintf(stderr,"Sending %d...",i);
#endif
    
    while ((u_int)endpt->tx_next == ((u_int)endpt->tx_first + *(endpt->tx_wall)))
      UNET_TRAP(hDriver, endpt->endpointId);
    
    *(endpt->tx_next) = txbd;
    
    UNET_TRAP(hDriver, endpt->endpointId);
    
#ifdef MDW_DEBUG
    fprintf(stderr,"sent!...");
#endif
    
    if (endpt->tx_next == endpt->tx_last)
      endpt->tx_next = endpt->tx_first;
    else
      endpt->tx_next++;
    
    while (endpt->rx_next->chan == UNET_CHANNEL_FREE) ;

    len = endpt->rx_next->length;
    num_bufs = (len / endpt->rxb_buffsize_used) + 1;

#ifdef MDW_DEBUG
    fprintf(stderr,"chan %d, len %d\n", endpt->rx_next->chan,
						endpt->rx_next->length);
    
    if (endpt->rx_next->length > endpt->info.rxd_payload_size) {

      for (n = 0; n < num_bufs; n++) {
        for (j = 0; j < MIN(endpt->rxb_buffsize_used, len); j++) {
	  fprintf(stderr,"%2.2x ",*((unsigned char *)(endpt->rx_next->pay_or_buf[n] + j)));
	}
	len -= MIN(endpt->rxb_buffsize_used, len);
      }
    } else {
	for (j = 0; j < endpt->rx_next->length; j++) {
	  fprintf(stderr,"%2.2x ",*((unsigned char *)((unsigned char *)endpt->rx_next->pay_or_buf + j)));
        }
    }
    fprintf(stderr,"got reply!\n");
#endif
    
    if (endpt->rx_next->length > endpt->info.rxd_payload_size) {
      for (n = 0; n < num_bufs; n++) {
        endpt->free_next->buf = (endpt->rx_next->pay_or_buf[n]);
        if (endpt->free_next == endpt->free_last)
	  endpt->free_next = endpt->free_first;
	else
	  endpt->free_next++;
      }
    }
    
    endpt->rx_next->chan = UNET_CHANNEL_FREE;
    *(endpt->rx_wall) = ((u_int)endpt->rx_next - (u_int)endpt->rx_first);
    if (endpt->rx_next == endpt->rx_last)
      endpt->rx_next = endpt->rx_first;
    else
      endpt->rx_next++;
    
  }

  t2 = get_seconds(); 

#ifndef MDW_TERSE_OUTPUT
  fprintf(stderr,"Sent %d %d-byte messages in %f seconds.\n",
	  num_messages,message_size,(t2-t1));
  fprintf(stderr,"%f usec/message or %f Mbit/sec\n",
	  ((t2-t1)*1e6)/num_messages,
	  (num_messages*message_size*8)/(1e6*(t2-t1)));
#else
	fprintf(stderr,"%d %d %f %f %f\n",
					num_messages,
					message_size,
					(t2-t1),
					((t2-t1)*1e6)/num_messages,
					(num_messages*message_size*8)/(1e6*(t2-t1)));
#endif

  CloseHandle(hDriver);
#ifndef MDW_TERSE_OUTPUT
  fprintf(stderr,"Closed, exiting\n");
#endif
  return 0;
}

